package eu.mhsl.minenet.minigames.instance.game.stateless.types.towerdefense;

import eu.mhsl.minenet.minigames.instance.Dimension;
import eu.mhsl.minenet.minigames.instance.game.stateless.types.towerdefense.generator.MazeGenerator;
import eu.mhsl.minenet.minigames.instance.game.stateless.types.towerdefense.towers.SkeletonTower;
import eu.mhsl.minenet.minigames.instance.game.stateless.types.towerdefense.towers.Tower;
import eu.mhsl.minenet.minigames.util.BatchUtil;
import eu.mhsl.minenet.minigames.util.CommonEventHandles;
import net.minestom.server.MinecraftServer;
import net.minestom.server.coordinate.Point;
import net.minestom.server.coordinate.Pos;
import net.minestom.server.entity.*;
import net.minestom.server.entity.attribute.Attribute;
import net.minestom.server.entity.damage.DamageType;
import net.minestom.server.event.entity.EntityDeathEvent;
import net.minestom.server.event.item.ItemDropEvent;
import net.minestom.server.event.player.PlayerTickEvent;
import net.minestom.server.event.player.PlayerUseItemEvent;
import net.minestom.server.instance.InstanceContainer;
import net.minestom.server.instance.block.Block;
import net.minestom.server.item.ItemStack;
import net.minestom.server.item.Material;

import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

public class TowerdefenseRoom extends InstanceContainer {
    private final Player player;
    private final Towerdefense game;
    private final List<EntityCreature> enemies = new ArrayList<>();
    private final List<Tower> towers = new ArrayList<>();
    private final Entity cursor;

    public TowerdefenseRoom(Player player, Towerdefense game) {
        super(UUID.randomUUID(), Dimension.OVERWORLD.key);
        MinecraftServer.getInstanceManager().registerInstance(this);
        this.player = player;
        this.game = game;
        this.player.setGameMode(GameMode.ADVENTURE);
        this.player.setAllowFlying(true);
        this.player.getAttribute(Attribute.BLOCK_INTERACTION_RANGE).setBaseValue(200);
        this.player.getInventory().setItemStack(0, ItemStack.of(Material.SKELETON_SPAWN_EGG));

        this.setGenerator(new MazeGenerator());
        BatchUtil.loadAndApplyBatch(this.game.getMazeBatch(), this, () -> {});

        this.cursor = new Entity(EntityType.ARMOR_STAND);
        this.cursor.setInstance(this);

        this.eventNode()
            .addListener(EntityDeathEvent.class, event -> {
                if(!(event.getEntity() instanceof EntityCreature enemy)) return;
                this.enemies.remove(enemy);
            })
            .addListener(PlayerTickEvent.class, this::setCursorPosition)
            .addListener(PlayerUseItemEvent.class, this::setTower)
            .addListener(ItemDropEvent.class, CommonEventHandles::cancel);
    }

    private void setTower(PlayerUseItemEvent event) {
        if(event.getItemStack().material().equals(Material.SKELETON_SPAWN_EGG)) {
            Point newPosition = this.player.getTargetBlockPosition(20);
            if(newPosition == null) return;
            if(!this.getBlock(newPosition).equals(Block.BLACK_WOOL)) return;
            this.setBlock(newPosition, Block.BLUE_WOOL);
            newPosition = newPosition.add(0.5,1,0.5);
            SkeletonTower tower = new SkeletonTower();
            tower.setInstance(this, newPosition);
        }
    }

    public List<EntityCreature> getEnemies() {
        return this.enemies;
    }

    private void setCursorPosition(PlayerTickEvent event) {
        Point newPosition = this.player.getTargetBlockPosition(20);
        if(newPosition == null) return;
        if(this.getBlock(newPosition).equals(Block.BLACK_WOOL)) {
            this.cursor.setInvisible(false);
        } else {
            this.cursor.setInvisible(true);
            return;
        }
        newPosition = newPosition.add(0.5,1,0.5);
        this.cursor.teleport(new Pos(newPosition));
    }

    void startWave(List<GroupFactory> enemyGroups) {
        enemyGroups.forEach(groupFactory -> groupFactory.summonGroup(this));
    }

    public void addEnemy(EntityCreature enemy) {
        enemy.setInstance(this, this.game.getMazePath().getFirst());
        this.enemies.add(enemy);
        this.changeEnemyGoal(enemy, 0);
    }

    public Towerdefense getGame() {
        return this.game;
    }

    private void changeEnemyGoal(EntityCreature enemy, int positionIndex) {
        if(positionIndex == this.game.getMazePath().size()-1) {
            this.enemies.remove(enemy);
            enemy.remove();
            this.player.damage(DamageType.PLAYER_ATTACK, enemy.getHealth()/10);
            return;
        }
        enemy.getNavigator().setPathTo(this.game.getMazePath().get(positionIndex+1), 0.6, () -> this.changeEnemyGoal(enemy, positionIndex+1));
    }
}
